import binascii
import datetime
import locale
import mimetypes
import os
import re
import subprocess
import time
import warnings
from typing import Union

try:
    import chardet
except ImportError:
    chardet = None  # type: ignore

from humanize import naturaltime
from werkzeug.middleware.proxy_fix import ProxyFix as WerkzeugProxyFix


class ProxyFix(WerkzeugProxyFix):
    """This middleware can be applied to add HTTP (reverse) proxy support to a
    WSGI application (klaus), making it possible to:

    * Mount it under a sub-URL (http://example.com/git/...)
    * Use a different HTTP scheme (HTTP vs. HTTPS)
    * Make it appear under a different domain altogether

    It sets `REMOTE_ADDR`, `HTTP_HOST` and `wsgi.url_scheme` from `X-Forwarded-*`
    headers.  It also sets `SCRIPT_NAME` from the `X-Script-Name` header.

    For instance if you have klaus mounted under /git/ and your site uses SSL
    (but your proxy doesn't), make the proxy pass ::

        X-Script-Name = '/git'
        X-Forwarded-Proto = 'https'
        ...

    If you have more than one proxy server in front of your app, set
    `num_proxies` accordingly.

    Do not use this middleware in non-proxy setups for security reasons.

    The original values of `REMOTE_ADDR` and `HTTP_HOST` are stored in
    the WSGI environment as `werkzeug.proxy_fix.orig_remote_addr` and
    `werkzeug.proxy_fix.orig_http_host`.

    :param app: the WSGI application
    :param num_proxies: the number of proxy servers in front of the app.
    """

    def __call__(self, environ, start_response):
        script_name = environ.get("HTTP_X_SCRIPT_NAME")
        if script_name is not None:
            if script_name.endswith("/"):
                warnings.warn(
                    "'X-Script-Name' header should not end in '/' (found: %r). "
                    "Please fix your proxy's configuration." % script_name
                )
                script_name = script_name.rstrip("/")
            environ["SCRIPT_NAME"] = script_name
        return super().__call__(environ, start_response)


class SubUri:
    """WSGI middleware to tweak the WSGI environ so that it's possible to serve
    the wrapped app (klaus) under a sub-URL and/or to use a different HTTP
    scheme (http:// vs. https://) for proxy communication.

    This is done by making your proxy pass appropriate HTTP_X_SCRIPT_NAME and
    HTTP_X_SCHEME headers.

    For instance if you have klaus mounted under /git/ and your site uses SSL
    (but your proxy doesn't), make it pass ::

        X-Script-Name = '/git'
        X-Scheme = 'https'

    Snippet stolen from http://flask.pocoo.org/snippets/35/
    """

    def __init__(self, app):
        warnings.warn(
            "'klaus.utils.SubUri' is deprecated and will be removed. "
            "Please upgrade your code to use 'klaus.utils.ProxyFix' instead.",
            DeprecationWarning,
        )
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get("HTTP_X_SCRIPT_NAME", "")
        if script_name:
            environ["SCRIPT_NAME"] = script_name.rstrip("/")

        if script_name and environ["PATH_INFO"].startswith(script_name):
            # strip `script_name` from PATH_INFO
            environ["PATH_INFO"] = environ["PATH_INFO"][len(script_name) :]

        if "HTTP_X_SCHEME" in environ:
            environ["wsgi.url_scheme"] = environ["HTTP_X_SCHEME"]

        return self.app(environ, start_response)


def timesince(when, now=time.time):
    """Return the difference between `when` and `now` in human readable form."""
    return naturaltime(now() - when)


def formattimestamp(timestamp):
    return datetime.datetime.fromtimestamp(timestamp).strftime("%b %d, %Y %H:%M:%S")


def guess_is_binary(dulwich_blob):
    return any(b"\0" in chunk for chunk in dulwich_blob.chunked)


def guess_is_image(filename):
    mime, _ = mimetypes.guess_type(filename)
    if mime is None:
        return False
    return mime.startswith("image/")


def encode_for_git(s):
    # XXX This assumes everything to be UTF-8 encoded
    return s.encode("utf8")


def decode_from_git(b):
    # XXX This assumes everything to be UTF-8 encoded
    return b.decode("utf8")


def force_unicode(s: Union[str, bytes]) -> str:
    """Do all kinds of magic to turn `s` into unicode"""
    # It's already unicode, don't do anything:
    if isinstance(s, str):
        return s

    # Try some default encodings:
    try:
        return s.decode("utf-8")
    except UnicodeDecodeError:
        pass

    try:
        return s.decode(locale.getpreferredencoding())
    except UnicodeDecodeError:
        pass

    if chardet is not None:
        # Try chardet, if available
        encoding = chardet.detect(s)["encoding"]
        if encoding is not None:
            try:
                return s.decode(encoding, "replace")
            except LookupError:
                pass

    return s.decode("latin1", "replace")


def extract_author_name(email):
    """Extract the name from an email address --
    >>> extract_author_name("John <john@example.com>")
    "John"

    -- or return the address if none is given.
    >>> extract_author_name("noname@example.com")
    "noname@example.com"
    """
    match = re.match("^(.*?)<.*?>$", email)
    if match:
        return match.group(1).strip()
    return email


def is_hex_prefix(s):
    if len(s) % 2:
        s += "0"
    try:
        binascii.unhexlify(s)
        return True
    except binascii.Error:
        return False


def shorten_sha1(sha1):
    if 20 <= len(sha1) <= 40 and is_hex_prefix(sha1):
        sha1 = sha1[:7]
    return sha1


def parent_directory(path):
    return os.path.split(path)[0]


def subpaths(path):
    """Yield a `(last part, subpath)` tuple for all possible sub-paths of `path`.

    >>> list(subpaths("foo/bar/spam"))
    [('foo', 'foo'), ('bar', 'foo/bar'), ('spam', 'foo/bar/spam')]
    """
    seen = []
    for part in path.split("/"):
        seen.append(part)
        yield part, "/".join(seen)


def shorten_message(msg):
    return msg.split("\n")[0]


def replace_dupes(ls, replacement):
    """Replace items in `ls` that are equal to their predecessors with `replacement`.

    >>> ls = [1, 2, 2, 3, 2, 2, 2]
    >>> replace_dupes(x, 'x')
    >>> ls
    [1, 2, 'x', 3, 2, 'x', 'x']
    """
    last = object()
    for i, elem in enumerate(ls):
        if last == elem:
            ls[i] = replacement
        else:
            last = elem


def guess_git_revision():
    """Try to guess whether this instance of klaus is run directly from a klaus
    git checkout.  If it is, guess and return the currently checked-out commit
    SHA.  If it's not (installed using pip, setup.py or the like), return None.

    This is used to display the "powered by klaus $VERSION" footer on each page,
    $VERSION being either the SHA guessed by this function or the latest release number.
    """
    git_dir = os.path.join(os.path.dirname(__file__), "..", ".git")
    try:
        return force_unicode(
            subprocess.check_output(
                ["git", "log", "--format=%h", "-n", "1"], cwd=git_dir
            ).strip()
        )
    except OSError:
        # Either the git executable couldn't be found in the OS's PATH
        # or no ".git" directory exists, i.e. this is no "bleeding-edge" installation.
        return None


def sanitize_branch_name(name, chars="./", repl="-"):
    for char in chars:
        name = name.replace(char, repl)
    return name


def escape_html(s):
    return (
        s.replace(b"&", b"&amp;")
        .replace(b"<", b"&lt;")
        .replace(b">", b"&gt;")
        .replace(b'"', b"&quot;")
    )


def tarball_basename(repo_name, rev):
    """Determine the name for a tarball."""
    rev = sanitize_branch_name(rev, chars="/")
    if rev.startswith(repo_name + "-"):
        # If the rev is a tag name that already starts with the repo name,
        # skip it.
        return rev
    elif len(rev) >= 2 and rev[0] == "v" and not rev[1].isalpha():
        # If the rev is a tag name prefixed by a 'v', skip the 'v'.
        # So, v-1.0 -> 1.0, v1.0 -> 1.0, but vanilla -> vanilla.
        return f"{repo_name}-{rev[1:]}"
    elif len(rev) == 40 and is_hex_prefix(rev):
        # If the rev is a commit hash, simply use that.
        return f"{repo_name}@{rev}"
    else:
        return f"{repo_name}-{rev}"


def repo_human_name(path):
    """Get repository name from path.

    1. /x/y.git -> /x/y  and  /x/y/.git/ -> /x/y//
    2. /x/y/ -> /x/y
    3. /x/y -> y
    """
    name = path.rstrip(os.sep).split(os.sep)[-1]
    if name.endswith(".git"):
        name = name[:-4]
    return name



import io
import os
import warnings

from .app_args import get_args_from_env
from .wsgi_autoreloading import make_autoreloading_app

if "KLAUS_REPOS" in os.environ:
    warnings.warn(
        "use KLAUS_REPOS_ROOT instead of KLAUS_REPOS for the autoreloader apps",
        DeprecationWarning,
    )

args, kwargs = get_args_from_env()
repos_root = os.environ.get("KLAUS_REPOS_ROOT") or os.environ["KLAUS_REPOS"]
args = (repos_root,) + args[1:]

if kwargs["htdigest_file"]:
    # Cache the contents of the htdigest file, the application will not read
    # the file like object until later when called.
    with open(kwargs["htdigest_file"], encoding="utf-8") as htdigest_file:
        kwargs["htdigest_file"] = io.StringIO(htdigest_file.read())

application = make_autoreloading_app(*args, **kwargs)



"""A cache for tagsfiles generated by the 'ctags' command line tool.

We don't want to run the 'ctags' command line tool on each request as it may
take a lot of time.  The following steps are necessary in order to create a
ctags tagsfile that be read by Pygments:

1. Clone the repository to a temporary location and check out the branch/commit
   the user is browsing, unless the branch is already checked out. (*)
2. Run 'ctags -R' on the temporary repository checkout.
3. Delete the temporary repository checkout.

To avoid going through these steps on each request, we cache the tagsfile
generated in step 2.  The cache is on-disk and non-persistent, i.e. cleared
whenever the Python interpreter running klaus is shut down.

For large projects, the ctags tagsfiles may grow to sizes of multiple MiB, so
we have to set an upper limit on the size of the cache.  Since tagsfiles are
represented as uncompressed ASCII files, we can increase the number of tagsfiles
we can cache by using compression.  Of course, 'python-ctags', which is used by
Pygments to read the tagsfiles, can't deal with compressed tagsfiles, so we have
to uncompress them before actually using them. To avoid decompressing tagsfiles
on each request, we keep the tagsfiles that are most likely to be used (**) in
uncompressed form.

(*) We always create a clone in the current implementation;
    this could be optimized in the future.
(**) "most likely": currently implemented as "most recently used"
"""
import atexit
import gzip
import os
import shutil
import tempfile
import threading

from dulwich.lru_cache import LRUSizeCache

from klaus.ctagsutils import create_tagsfile, delete_tagsfile

# Good compression while taking only 10% more time than level 1
COMPRESSION_LEVEL = 4


def compress_tagsfile(uncompressed_tagsfile_path):
    """Compress an uncompressed tagsfile.

    :return: path to the compressed version of the tagsfile
    """
    _, compressed_tagsfile_path = tempfile.mkstemp()
    with open(uncompressed_tagsfile_path, "rb") as uncompressed:
        with gzip.open(compressed_tagsfile_path, "wb", COMPRESSION_LEVEL) as compressed:
            shutil.copyfileobj(uncompressed, compressed)
    return compressed_tagsfile_path


def uncompress_tagsfile(compressed_tagsfile_path):
    """Uncompress an compressed tagsfile.

    :return: path to the uncompressed version of the tagsfile
    """
    _, uncompressed_tagsfile_path = tempfile.mkstemp()
    with gzip.open(compressed_tagsfile_path, "rb") as compressed:
        with open(uncompressed_tagsfile_path, "wb") as uncompressed:
            shutil.copyfileobj(compressed, uncompressed)
    return uncompressed_tagsfile_path


MiB = 1024 * 1024


class CTagsCache:
    """A ctags cache. Both uncompressed and compressed entries are kept in
    temporary files created by `tempfile.mkstemp` which are deleted from disk
    when the Python interpreter is shut down.

    :param uncompressed_max_bytes: Maximum size of the uncompressed cache sector
    :param compressed_max_bytes:   Maximum size of the compressed cache sector

    The lifecycle of a cache entry is as follows.

    - When first created, a tagsfile is put into the uncompressed cache sector.
    - When free space is required for other uncompressed tagsfiles, it may be
      moved to the compressed cache sector. Gzip is used to compress the tagsfile.
    - When free space is required for other compressed tagsfiles, it may be
      evicted from the cache entirely.
    - When the tagsfile is requested and it's in the compressed cache sector,
      it is moved back to the uncompressed sector prior to using it.
    """

    def __init__(self, uncompressed_max_bytes=30 * MiB, compressed_max_bytes=20 * MiB):
        self.uncompressed_max_bytes = uncompressed_max_bytes
        self.compressed_max_bytes = compressed_max_bytes
        # Note: We use dulwich's LRU cache to store the tagsfile paths here,
        # but we could easily replace it by any other (LRU) cache implementation.
        self._uncompressed_cache = LRUSizeCache(
            uncompressed_max_bytes, compute_size=os.path.getsize
        )
        self._compressed_cache = LRUSizeCache(
            compressed_max_bytes, compute_size=os.path.getsize
        )
        self._clearing = False
        self._lock = threading.Lock()

        atexit.register(self.clear)

    def __del__(self):
        self.clear()

    def clear(self):
        """Clear both the uncompressed and compressed caches."""
        # Don't waste time moving tagsfiles from uncompressed to compressed cache,
        # but remove them directly instead:
        self._clearing = True
        self._uncompressed_cache.clear()
        self._compressed_cache.clear()
        self._clearing = False

    def get_tagsfile(self, git_repo_path, git_rev):
        """Get the ctags tagsfile for the given Git repository and revision.

        - If the tagsfile is still in cache, and in uncompressed form, return it
          without any further cost.
        - If the tagsfile is still in cache, but in compressed form, uncompress
          it, put it into uncompressed space, and return the uncompressed version.
        - If the tagsfile isn't in cache at all, create it, put it into
          uncompressed cache and return the newly created version.
        """
        # Always require full SHAs
        assert len(git_rev) == 40

        # Avoiding race conditions, The Sledgehammer Way
        with self._lock:
            if git_rev in self._uncompressed_cache:
                return self._uncompressed_cache[git_rev]

            if git_rev in self._compressed_cache:
                compressed_tagsfile_path = self._compressed_cache[git_rev]
                uncompressed_tagsfile_path = uncompress_tagsfile(
                    compressed_tagsfile_path
                )
                self._compressed_cache._remove_node(
                    self._compressed_cache._cache[git_rev]
                )
            else:
                # Not in cache.
                uncompressed_tagsfile_path = create_tagsfile(git_repo_path, git_rev)
            self._uncompressed_cache.add(
                git_rev, uncompressed_tagsfile_path, self._clear_uncompressed_entry
            )
            return uncompressed_tagsfile_path

    def _clear_uncompressed_entry(self, git_rev, uncompressed_tagsfile_path):
        """Called by LRUSizeCache whenever an entry is to be evicted from
        uncompressed cache.

        Most of the times this happens when space is needed
        in uncompressed cache, in which case we move the tagsfile to compressed
        cache.  When clearing the cache, we don't bother moving entries to
        uncompressed space; we delete them directly instead.
        """
        if not self._clearing:
            # If we're clearing the whole cache, don't waste time moving tagsfiles
            # from uncompressed to compressed cache, but remove them directly instead.
            self._compressed_cache.add(
                git_rev,
                compress_tagsfile(uncompressed_tagsfile_path),
                self._clear_compressed_entry,
            )
        delete_tagsfile(uncompressed_tagsfile_path)

    def _clear_compressed_entry(self, git_rev, compressed_tagsfile_path):
        """Called by LRUSizeCache whenever an entry to be evicted from
        compressed cache.

        This happens when space is needed for new compressed
        tagsfiles.  We delete the evictee from the cache entirely.
        """
        delete_tagsfile(compressed_tagsfile_path)



import re
import shutil
import subprocess
import tempfile

import pytest
import requests
import requests.auth

import klaus

from .utils import *


def test_make_app_using_list():
    app = klaus.make_app(REPOS, TEST_SITE_NAME)
    with serve_app(app):
        response = requests.get(UNAUTH_TEST_SERVER).text
        assert TEST_REPO_NO_NEWLINE_BASE_URL in response


def test_htdigest_file_without_smarthttp_or_require_browser_auth():
    with pytest.raises(ValueError):
        klaus.make_app([], None, htdigest_file=object())


def test_unauthenticated_push_and_require_browser_auth():
    with pytest.raises(ValueError):
        klaus.make_app(
            [],
            None,
            use_smarthttp=True,
            unauthenticated_push=True,
            require_browser_auth=True,
        )


def test_unauthenticated_push_without_use_smarthttp():
    with pytest.raises(ValueError):
        klaus.make_app([], None, unauthenticated_push=True)


def test_unauthenticated_push_with_disable_push():
    with pytest.raises(ValueError):
        klaus.make_app([], None, unauthenticated_push=True, disable_push=True)


def options_test(make_app_args, expected_permissions):
    def test():
        with serve(**make_app_args):
            for check, permitted in expected_permissions.items():
                if check in globals():
                    checks = [check]
                elif check.endswith("auth"):
                    checks = ["can_%s" % check]
                else:
                    checks = ["can_%s_unauth" % check, "can_%s_auth" % check]
                for check in checks:
                    assert globals()[check]() == permitted, check

    return test


test_nosmart_noauth = options_test({}, {"reach": True, "clone": False, "push": False})
test_smart_noauth = options_test(
    {"use_smarthttp": True}, {"reach": True, "clone": True, "push": False}
)
test_smart_push = options_test(
    {"use_smarthttp": True, "htdigest_file": open(HTDIGEST_FILE)},
    {"reach": True, "clone": True, "push_auth": True, "push_unauth": False},
)
test_unauthenticated_push = options_test(
    {"use_smarthttp": True, "unauthenticated_push": True},
    {"reach": True, "clone": True, "push": True},
)
test_nosmart_auth = options_test(
    {"require_browser_auth": True, "htdigest_file": open(HTDIGEST_FILE)},
    {"reach_auth": True, "reach_unauth": False, "clone": False, "push": False},
)
test_smart_auth = options_test(
    {
        "require_browser_auth": True,
        "use_smarthttp": True,
        "htdigest_file": open(HTDIGEST_FILE),
    },
    {
        "reach_auth": True,
        "reach_unauth": False,
        "clone_auth": True,
        "clone_unauth": False,
        "push_unauth": False,
        "push_auth": True,
    },
)
test_smart_auth_disable_push = options_test(
    {
        "require_browser_auth": True,
        "use_smarthttp": True,
        "disable_push": True,
        "htdigest_file": open(HTDIGEST_FILE),
    },
    {
        "reach_auth": True,
        "reach_unauth": False,
        "clone_auth": True,
        "clone_unauth": False,
        "push": False,
    },
)

test_ctags_disabled = options_test(
    {}, {"ctags_tags_and_branches": False, "ctags_all": False}
)
test_ctags_tags_and_branches = options_test(
    {"ctags_policy": "tags-and-branches"},
    {"ctags_tags_and_branches": True, "ctags_all": False},
)
test_ctags_all = options_test(
    {"ctags_policy": "ALL"}, {"ctags_tags_and_branches": True, "ctags_all": True}
)


# Reach
def can_reach_unauth():
    return _check_http200(_GET_unauth, TEST_REPO_BASE_URL)


def can_reach_auth():
    return _check_http200(_GET_auth, TEST_REPO_BASE_URL)


# Clone
def can_clone_unauth():
    return _can_clone(_GET_unauth, UNAUTH_TEST_REPO_URL)


def can_clone_auth():
    return _can_clone(_GET_auth, AUTH_TEST_REPO_URL)


def _can_clone(http_get, url):
    tmp = tempfile.mkdtemp()
    try:
        return all(
            [
                "git clone" in http_get(TEST_REPO_BASE_URL).text,
                _check_http200(
                    http_get, TEST_REPO_BASE_URL + "info/refs?service=git-upload-pack"
                ),
                subprocess.call(["git", "clone", url, tmp]) == 0,
            ]
        )
    finally:
        shutil.rmtree(tmp, ignore_errors=True)


# Push
def can_push_unauth():
    return _can_push(_GET_unauth, UNAUTH_TEST_REPO_URL)


def can_push_auth():
    return _can_push(_GET_auth, AUTH_TEST_REPO_URL)


def _can_push(http_get, url):
    return all(
        [
            any(
                [
                    _check_http200(
                        http_get,
                        TEST_REPO_BASE_URL + "info/refs?service=git-receive-pack",
                    ),
                    _check_http200(http_get, TEST_REPO_BASE_URL + "git-receive-pack"),
                ]
            ),
            subprocess.call(["git", "push", url, "master"], cwd=TEST_REPO) == 0,
        ]
    )


# Ctags
def ctags_tags_and_branches():
    return all(
        _ctags_enabled(ref, f)
        for ref in ["master", "tag1"]
        for f in ["test.c", "test.js"]
    )


def ctags_all():
    all_refs = re.findall(
        'href=".+/commit/([a-z0-9]{40})/">', requests.get(UNAUTH_TEST_REPO_URL).text
    )
    assert len(all_refs) == 3
    return all(
        _ctags_enabled(ref, f) for ref in all_refs for f in ["test.c", "test.js"]
    )


def _ctags_enabled(ref, filename):
    response = requests.get(UNAUTH_TEST_REPO_URL + f"blob/{ref}/{filename}")
    assert response.status_code == 200, response.text
    href = f'<a href="/{TEST_REPO_BASE_URL}blob/{ref}/{filename}#L-1">'
    return href in response.text


def _GET_unauth(url=""):
    return requests.get(
        UNAUTH_TEST_SERVER + url,
        auth=requests.auth.HTTPDigestAuth("invalid", "password"),
    )


def _GET_auth(url=""):
    return requests.get(
        AUTH_TEST_SERVER + url,
        auth=requests.auth.HTTPDigestAuth("testuser", "testpassword"),
    )


def _check_http200(http_get, url):
    return http_get(url).status_code == 200



import requests

from .utils import *


def test_blame():
    with serve():
        response = requests.get(UNAUTH_TEST_REPO_URL + "blame/HEAD/test.c")
        assert response.status_code == 200


def test_blame_empty():
    with serve():
        response = requests.get(UNAUTH_TEST_REPO_URL + "blame/HEAD/empty.txt")
        assert response.status_code == 200


def test_dont_show_blame_link():
    with serve():
        for file in ["binary", "image.jpg", "toolarge"]:
            response = requests.get(
                UNAUTH_TEST_REPO_DONT_RENDER_URL + "blob/HEAD/" + file
            ).text
            assert "blame" not in response


def test_dont_render_blame():
    """Don't render blame even if someone navigated to the blame site by accident."""
    with serve():
        for file in ["binary", "image.jpg", "toolarge"]:
            response = requests.get(
                UNAUTH_TEST_REPO_DONT_RENDER_URL + "blame/HEAD/" + file
            ).text
            assert "Can't show blame" in response



import glob
import threading
import time

from klaus import make_app


# Shared state between poller and application wrapper
class S:
    #: the real WSGI app
    inner_app = None
    should_reload = True


def poll_for_changes(interval, dir):
    """
    Polls `dir` for changes every `interval` seconds and sets `should_reload`
    accordingly.
    """
    glob_pattern = dir + "/*"
    old_contents = glob.glob(glob_pattern)
    while 1:
        time.sleep(interval)
        if S.should_reload:
            # klaus application has not seen our change yet
            continue
        new_contents = glob.glob(glob_pattern)
        if new_contents != old_contents:
            # Directory contents changed => should_reload
            old_contents = new_contents
            S.should_reload = True


def make_autoreloading_app(repos_root, *args, **kwargs):
    def app(environ, start_response):
        if S.should_reload:
            # Refresh inner application with new repo list
            print("Reloading repository list...")
            S.inner_app = make_app(glob.glob(repos_root + "/*"), *args, **kwargs)
            S.should_reload = False
        return S.inner_app(environ, start_response)

    # Background thread that polls the directory for changes
    poller_thread = threading.Thread(target=(lambda: poll_for_changes(10, repos_root)))
    poller_thread.daemon = True
    poller_thread.start()

    return app



import sys
import unittest
from unittest import mock

from klaus import utils


class ForceUnicodeTests(unittest.TestCase):
    def test_ascii(self):
        self.assertEqual("foo", utils.force_unicode(b"foo"))

    def test_utf8(self):
        if sys.version_info[0] < 3:
            return
        self.assertEqual(eval(r'"f\xce"'), utils.force_unicode(eval(r'b"f\xc3\x8e"')))

    def test_invalid(self):
        if sys.platform.startswith("win") or sys.version_info[0] < 3:
            return
        with mock.patch.object(utils, "chardet", None):
            self.assertEqual(b"f\xc3\x8e", utils.force_unicode(b"f\xce").encode("utf8"))


class TarballBasenameTests(unittest.TestCase):
    def test_examples(self):
        examples = [
            ("v0.1", "klaus-0.1"),
            ("klaus-0.1", "klaus-0.1"),
            ("0.1", "klaus-0.1"),
            (
                "b3e70e08344ca3f83cc7033ecdbefa90443d7d2e",
                "klaus@b3e70e08344ca3f83cc7033ecdbefa90443d7d2e",
            ),
            ("vanilla", "klaus-vanilla"),
        ]
        for rev, basename in examples:
            self.assertEqual(utils.tarball_basename("klaus", rev), basename)



import functools
import io
import os
import stat
import subprocess
import threading

import dulwich
import dulwich.patch
from dulwich.errors import NotTreeError
from dulwich.object_store import tree_lookup_path
from dulwich.objects import S_ISGITLINK, Blob

try:
    from dulwich.refs import SymrefLoop
except ImportError:  # dulwich < 0.20.46
    InaccessibleRef = KeyError  # type: ignore
else:
    InaccessibleRef = (SymrefLoop, KeyError)  # type: ignore


from klaus.diff import render_diff
from klaus.utils import (
    decode_from_git,
    encode_for_git,
    force_unicode,
    parent_directory,
    repo_human_name,
)

NOT_SET = "__not_set__"


def cached_call(key, validator, producer, _cache={}):
    data, old_validator = _cache.get(key, (None, NOT_SET))
    if old_validator != validator:
        data = producer()
    _cache[key] = (data, validator)
    return data


def synchronized(func, lock=threading.RLock()):
    @functools.wraps(func)
    def synchronized_func(*args, **kwargs):
        with lock:
            return func(*args, **kwargs)

    return synchronized_func


class FancyRepo:
    """A wrapper around Dulwich's Repo that adds some helper methods."""

    def __init__(self, path, namespace):
        self.dulwich_repo = dulwich.repo.Repo(path)
        self.namespace = namespace

    @synchronized
    def __getattr__(self, attr):
        return getattr(self.dulwich_repo, attr)

    @synchronized
    def __getitem__(self, key):
        return self.dulwich_repo[key]

    @property
    def name(self):
        return repo_human_name(self.path)

    @property
    def namespaced_name(self):
        if self.namespace:
            return f"~{self.namespace}/{self.name}"
        else:
            return self.name

    # TODO: factor out stuff into dulwich
    @synchronized
    def get_last_updated_at(self):
        """Get datetime of last commit to this repository.

        Caches the result to speed up the repo_list page.
        Cache is invalidated if one of the ref targets changes,
        eg. a new commit has been made and 'refs/heads/master' was changed.
        """

        def _get_commit_time_cached(ref_id):
            return cached_call(
                key=(ref_id, "_get_commit_time"),
                validator=None,
                producer=lambda: _get_commit_time(ref_id),
            )

        def _get_commit_time(ref_id):
            try:
                return self[ref_id].commit_time
            except (KeyError, AttributeError):
                # Missing or non-commit object
                return None

        max_refs = 1000
        if len(self.dulwich_repo.refs.keys()) > max_refs:
            # If we have too many refs, look at the branches only. (And HEAD, see below.)
            base = b"refs/heads"
        else:
            base = None
        all_ids = list(self.dulwich_repo.refs.as_dict(base).values())
        # If we still have too many refs, keep only some.
        if len(all_ids) > max_refs:
            all_ids = sorted(all_ids)[:max_refs]
        # Always add HEAD.
        try:
            all_ids.append(self.dulwich_repo.refs[b"HEAD"])
        except KeyError:
            pass

        commit_times = filter(None, map(_get_commit_time_cached, all_ids))
        try:
            return max(commit_times)
        except ValueError:
            # Python 2 does not support max(..., default=)
            return None

    @property
    @synchronized
    def cloneurl(self):
        """Retrieve the gitweb notion of the public clone URL of this repo."""
        f = self.dulwich_repo.get_named_file("cloneurl")
        if f is not None:
            return force_unicode(f.read())
        c = self.dulwich_repo.get_config()
        try:
            return force_unicode(c.get(b"gitweb", b"url"))
        except KeyError:
            return None

    @synchronized
    def get_description(self):
        """Like Dulwich's `get_description`, but returns None if the file
        contains Git's default text "Unnamed repository[...]".
        """
        # Cache result to speed up repo_list.html template.
        # If description file mtime has changed, we should invalidate the cache.
        description_file = os.path.join(self.dulwich_repo._controldir, "description")
        try:
            description_mtime = os.stat(description_file).st_mtime
        except OSError:
            description_mtime = None

        return cached_call(
            key=(id(self), "get_description"),
            validator=description_mtime,
            producer=self._get_description,
        )

    def _get_description(self):
        description = self.dulwich_repo.get_description()
        if description:
            description = force_unicode(description)
            if not description.startswith("Unnamed repository;"):
                return force_unicode(description)

    @synchronized
    def get_commit(self, rev):
        """Get commit object identified by `rev` (SHA or branch or tag name)."""
        for prefix in ["refs/heads/", "refs/tags/", ""]:
            key = prefix + rev
            try:
                obj = self[encode_for_git(key)]
                if isinstance(obj, dulwich.objects.Tag):
                    obj = self[obj.object[1]]
                return obj
            except KeyError:
                pass
        raise KeyError(rev)

    @synchronized
    def get_default_branch(self):
        """Tries to guess the default repo branch name."""
        for candidate in ["master", "main", "trunk", "default", "gh-pages"]:
            try:
                self.get_commit(candidate)
                return candidate
            except InaccessibleRef:
                pass
        return None

    @synchronized
    def get_ref_names_ordered_by_last_commit(self, prefix, exclude=None):
        """Return a list of ref names that begin with `prefix`, ordered by the
        time they have been committed to last.
        """

        def get_commit_time(refname):
            try:
                obj = self[refs[refname]]
            except InaccessibleRef:
                # Default to 0, i.e. sorting refs that point at non-existant
                # objects last.
                return 0
            if isinstance(obj, dulwich.objects.Tag):
                return obj.tag_time
            return obj.commit_time

        refs = self.dulwich_repo.refs.as_dict(encode_for_git(prefix))
        if exclude:
            refs.pop(prefix + exclude, None)
        sorted_names = sorted(refs.keys(), key=get_commit_time, reverse=True)
        return [decode_from_git(ref) for ref in sorted_names]

    @synchronized
    def get_branch_names(self, exclude=None):
        """Return a list of branch names of this repo, ordered by the time they
        have been committed to last.
        """
        return self.get_ref_names_ordered_by_last_commit("refs/heads", exclude)

    @synchronized
    def get_tag_names(self):
        """Return a list of tag names of this repo, ordered by creation time."""
        return self.get_ref_names_ordered_by_last_commit("refs/tags")

    @synchronized
    def get_tag_and_branch_shas(self):
        """Return a list of SHAs of all tags and branches."""
        tag_shas = self.dulwich_repo.refs.as_dict(b"refs/tags/").values()
        branch_shas = self.dulwich_repo.refs.as_dict(b"refs/heads/").values()
        return set(tag_shas) | set(branch_shas)

    @synchronized
    def history(self, commit, path=None, max_commits=None, skip=0):
        """Return a list of all commits that affected `path`, starting at branch
        or commit `commit`. `skip` can be used for pagination, `max_commits`
        to limit the number of commits returned.

        Similar to `git log [branch/commit] [--skip skip] [-n max_commits]`.
        """
        # XXX The pure-Python/dulwich code is very slow compared to `git log`
        #     at the time of this writing (mid-2012).
        #     For instance, `git log .tx` in the Django root directory takes
        #     about 0.15s on my machine whereas the history() method needs 5s.
        #     Therefore we use `git log` here until dulwich gets faster.
        #     For the pure-Python implementation, see the 'purepy-hist' branch.

        cmd = ["git", "log", "--format=%H"]
        if skip:
            cmd.append("--skip=%d" % skip)
        if max_commits:
            cmd.append("--max-count=%d" % max_commits)
        cmd.append(decode_from_git(commit.id))
        if path:
            cmd.extend(["--", path])

        output = subprocess.check_output(cmd, cwd=os.path.abspath(self.path))
        sha1_sums = output.strip().split(b"\n")
        return [self[sha1] for sha1 in sha1_sums]

    @synchronized
    def blame(self, commit, path):
        """Return a 'git blame' list for the file at `path`: For each line in
        the file, the list contains the commit that last changed that line.
        """
        # XXX see comment in `.history()`
        cmd = ["git", "blame", "-ls", "--root", decode_from_git(commit.id), "--", path]
        output = subprocess.check_output(cmd, cwd=os.path.abspath(self.path))
        sha1_sums = [line[:40] for line in output.strip().split(b"\n") if line]
        return [
            None if self[sha1] is None else decode_from_git(self[sha1].id)
            for sha1 in sha1_sums
        ]

    @synchronized
    def get_blob_or_tree(self, commit, path):
        """Return the Git tree or blob object for `path` at `commit`."""
        try:
            (mode, oid) = tree_lookup_path(
                self.dulwich_repo.__getitem__, commit.tree, encode_for_git(path)
            )
        except NotTreeError:
            # Some part of the path was a file where a folder was expected.
            # Example: path="/path/to/foo.txt" but "to" is a file in "/path".
            raise KeyError
        return self[oid]

    @synchronized
    def listdir(self, commit, path):
        """Return a list of submodules, directories and files in given
        directory: Lists of (link name, target path) tuples.
        """
        submodules, dirs, files = [], [], []
        for entry_rel in self.get_blob_or_tree(commit, path).items():
            # entry_rel: Entry('foo.txt', ...)
            # entry_abs: Entry('spam/eggs/foo.txt', ...)
            entry_abs = entry_rel.in_path(encode_for_git(path))
            path_str = decode_from_git(entry_abs.path)
            item = (os.path.basename(path_str), path_str)
            if S_ISGITLINK(entry_abs.mode):
                submodules.append(item)
            elif stat.S_ISDIR(entry_abs.mode):
                dirs.append(item)
            else:
                files.append(item)

        def keyfunc(tpl):
            return tpl[0].lower()

        submodules.sort(key=keyfunc)
        files.sort(key=keyfunc)
        dirs.sort(key=keyfunc)

        if path:
            dirs.insert(0, ("..", parent_directory(path)))

        return {"submodules": submodules, "dirs": dirs, "files": files}

    @synchronized
    def commit_diff(self, commit):
        """Return the list of changes introduced by `commit`."""
        from klaus.utils import guess_is_binary

        if commit.parents:
            parent_tree = self[commit.parents[0]].tree
        else:
            parent_tree = None

        summary = {"nfiles": 0, "nadditions": 0, "ndeletions": 0}
        file_changes = []  # the changes in detail

        dulwich_changes = self.dulwich_repo.object_store.tree_changes(
            parent_tree, commit.tree
        )
        for (oldpath, newpath), (oldmode, newmode), (oldsha, newsha) in dulwich_changes:
            summary["nfiles"] += 1
            try:
                oldblob = (
                    self.dulwich_repo.object_store[oldsha]
                    if oldsha
                    else Blob.from_string(b"")
                )
            except KeyError:
                # probably related to submodules; Dulwich will handle that.
                oldblob = Blob.from_string(b"")
            try:
                newblob = (
                    self.dulwich_repo.object_store[newsha]
                    if newsha
                    else Blob.from_string(b"")
                )
            except KeyError:
                # probably related to submodules; Dulwich will handle that.
                newblob = Blob.from_string(b"")

            # Check for binary files -- can't show diffs for these
            if guess_is_binary(newblob) or guess_is_binary(oldblob):
                file_changes.append(
                    {
                        "is_binary": True,
                        "old_filename": oldpath or "/dev/null",
                        "new_filename": newpath or "/dev/null",
                        "chunks": None,
                    }
                )
                continue

            additions, deletions, chunks = render_diff(
                oldblob.splitlines(), newblob.splitlines()
            )
            change = {
                "is_binary": False,
                "old_filename": oldpath or "/dev/null",
                "new_filename": newpath or "/dev/null",
                "chunks": chunks,
                "additions": additions,
                "deletions": deletions,
            }
            summary["nadditions"] += additions
            summary["ndeletions"] += deletions
            file_changes.append(change)

        return summary, file_changes

    @synchronized
    def raw_commit_diff(self, commit):
        if commit.parents:
            parent_tree = self[commit.parents[0]].tree
        else:
            parent_tree = None
        bytesio = io.BytesIO()
        dulwich.patch.write_tree_diff(
            bytesio, self.dulwich_repo.object_store, parent_tree, commit.tree
        )
        return bytesio.getvalue()

    def freeze(self):
        return FrozenFancyRepo(self)


class FrozenFancyRepo:
    """A special version of FancyRepo that assumes the underlying Git
    repository does not change.  Used for performance optimizations.
    """

    def __init__(self, repo):
        self.__repo = repo
        self.__last_updated_at = NOT_SET

    def __setattr__(self, name, value):
        if not name.startswith("_FrozenFancyRepo__"):
            raise TypeError("Can't set %s attribute on FrozenFancyRepo" % name)
        super().__setattr__(name, value)

    def __getattr__(self, name):
        return getattr(self.__repo, name)

    def fast_get_last_updated_at(self):
        if self.__last_updated_at is NOT_SET:
            self.__last_updated_at = self.__repo.get_last_updated_at()
        return self.__last_updated_at


class InvalidRepo:
    """Represent an invalid repository and store pertinent data."""

    def __init__(self, path, namespace):
        self.path = path
        self.namespace = namespace

    @property
    def name(self):
        return repo_human_name(self.path)

    @property
    def namespaced_name(self):
        if self.namespace:
            return f"~{self.namespace}/{self.name}"
        else:
            return self.name



import os
import sys
from io import BytesIO

import dulwich.archive
import dulwich.config
import dulwich.objects
from dulwich.object_store import tree_lookup_path
from flask import current_app, render_template, request, url_for
from flask.views import View
from werkzeug.exceptions import NotFound
from werkzeug.wrappers import Response

try:
    from dulwich.refs import SymrefLoop
except ImportError:  # dulwich < 0.20.46

    class SymrefLoop(Exception):  # type: ignore
        """Dummy exception."""


try:
    import ctags
except ImportError:
    ctags = None
else:
    from klaus import ctagscache

    CTAGS_CACHE = ctagscache.CTagsCache()

from klaus import markup
from klaus.highlighting import highlight_or_render
from klaus.utils import (
    encode_for_git,
    force_unicode,
    guess_is_binary,
    guess_is_image,
    parent_directory,
    replace_dupes,
    sanitize_branch_name,
    subpaths,
)

README_FILENAMES = [
    b"README",
    b"README.md",
    b"README.mkdn",
    b"README.mdwn",
    b"README.markdown",
    b"README.rst",
]


def repo_list():
    """Show a list of all repos. Can be sorted by last update and repo names can be searched."""
    repos = [repo.freeze() for repo in current_app.valid_repos.values()]
    invalid_repos = current_app.invalid_repos.values()

    order_by = request.args.get("order_by") or "last_updated"
    search_query = request.args.get("q") or ""

    if search_query:
        repos = [r for r in repos if search_query.lower() in r.namespaced_name.lower()]
        invalid_repos = [
            r
            for r in invalid_repos
            if search_query.lower() in r.namespaced_name.lower()
        ]

    if order_by == "name":

        def sort_key(repo):
            return repo.namespaced_name

    else:

        def sort_key(repo):
            return -(repo.fast_get_last_updated_at() or -1), repo.namespaced_name

    repos = sorted(repos, key=sort_key)
    invalid_repos = sorted(invalid_repos, key=lambda repo: repo.namespaced_name)

    return render_template(
        "repo_list.html",
        repos=repos,
        invalid_repos=invalid_repos,
        order_by=order_by,
        search_query=search_query,
        base_href=None,
    )


def robots_txt():
    """Serve the robots.txt file to manage the indexing of the site by search engines."""
    return current_app.send_static_file("robots.txt")


def _get_repo_and_rev(repo, namespace=None, rev=None, path=None):
    if path and rev:
        rev += "/" + path.rstrip("/")

    if namespace:
        repo_key = f"~{namespace}/{repo}"
    else:
        repo_key = repo
    try:
        repo = current_app.valid_repos[repo_key]
    except KeyError:
        raise NotFound("No such repository %r" % repo)

    if rev is None:
        rev = repo.get_default_branch()
        if rev is None:
            rev = "HEAD"
            try:
                repo.get_commit("HEAD")
            except KeyError:
                raise NotFound("No commits yet")

    i = len(rev)
    while i > 0:
        try:
            commit = repo.get_commit(rev[:i])
            path = rev[i:].strip("/")
            rev = rev[:i]
        except (OSError, KeyError):
            i = rev.rfind("/", 0, i)
        except SymrefLoop as e:
            raise NotFound("symref loop for %s at depth %d" % (e.ref, e.depth))
        else:
            break
    else:
        raise NotFound("No such commit %r" % rev)

    return repo, rev, path, commit


def _get_submodule(repo, commit, path):
    """Retrieve submodule URL and path."""
    submodule_blob = repo.get_blob_or_tree(commit, ".gitmodules")
    config = dulwich.config.ConfigFile.from_file(
        BytesIO(submodule_blob.as_raw_string())
    )
    key = (b"submodule", path)
    submodule_url = config.get(key, b"url")
    submodule_path = config.get(key, b"path")
    return (submodule_url, submodule_path)


class BaseRepoView(View):
    """Base for all views with a repo context.

    The arguments `repo`, `rev`, `path` (see `dispatch_request`) define the
    repository, branch/commit and directory/file context, respectively --
    that is, they specify what (and in what state) is being displayed in all the
    derived views.

    For example: The 'history' view is the `git log` equivalent, i.e. if `path`
    is "/foo/bar", only commits related to "/foo/bar" are displayed, and if
    `rev` is "master", the history of the "master" branch is displayed.
    """

    def __init__(self, view_name):
        self.view_name = view_name
        self.context = {}

    def dispatch_request(self, repo, namespace=None, rev=None, path=""):
        """Dispatch repository, revision (if any) and path (if any). To retain
        compatibility with :func:`url_for`, view routing uses two arguments:
        rev and path, although a single path is sufficient (from Git's point of
        view, '/foo/bar/baz' may be a branch '/foo/bar' containing baz, or a
        branch '/foo' containing 'bar/baz', but never both [1].

        Hence, rebuild rev and path to a single path argument, which is then
        later split into rev and path again, but revision now may contain
        slashes.

        [1] https://github.com/jonashaag/klaus/issues/36#issuecomment-23990266
        """
        self.make_template_context(repo, namespace, rev, path.strip("/"))
        return self.get_response()

    def get_response(self):
        return render_template(self.template_name, **self.context)

    def make_template_context(self, repo, namespace, rev, path):
        repo, rev, path, commit = _get_repo_and_rev(repo, namespace, rev, path)

        try:
            blob_or_tree = repo.get_blob_or_tree(commit, path)
        except KeyError:
            raise NotFound("File not found")

        self.context = {
            "view": self.view_name,
            "repo": repo,
            "namespace": namespace,
            "rev": rev,
            "commit": commit,
            "branches": repo.get_branch_names(exclude=rev),
            "tags": repo.get_tag_names(),
            "path": path,
            "blob_or_tree": blob_or_tree,
            "subpaths": list(subpaths(path)) if path else None,
            "base_href": None,
        }


class CommitView(BaseRepoView):
    template_name = "view_commit.html"


class PatchView(BaseRepoView):
    def get_response(self):
        return Response(
            self.context["repo"].raw_commit_diff(self.context["commit"]),
            mimetype="text/plain",
        )


class TreeViewMixin:
    """The logic required for displaying the current directory in the sidebar."""

    def make_template_context(self, *args):
        super().make_template_context(*args)
        self.context["root_tree"] = self.listdir()

    def listdir(self):
        """Return a list of directories and files in the current path of the selected commit."""
        root_directory = self.get_root_directory()
        return self.context["repo"].listdir(self.context["commit"], root_directory)

    def get_root_directory(self):
        root_directory = self.context["path"]
        if isinstance(self.context["blob_or_tree"], dulwich.objects.Blob):
            # 'path' is a file (not folder) name
            root_directory = parent_directory(root_directory)
        return root_directory


class ReadmeMixin:
    """The logic required for finding and displaying README files."""

    def _get_readme(self):
        commit, path = self.context["commit"], self.context["path"]
        tree = self.context["repo"].get_blob_or_tree(commit, path)

        if not isinstance(tree, dulwich.objects.Tree):
            raise KeyError

        for name in README_FILENAMES:
            if name.lower() in [t.lower() for t in tree]:
                obj = self.context["repo"][tree[name][1]]
                if obj.type_name == b"blob":
                    readme_data = obj.data
                    readme_filename = name
                    return (readme_filename, readme_data)
        else:
            raise KeyError

    def get_readme_context(self):
        try:
            (readme_filename, readme_data) = self._get_readme()
        except KeyError:
            return {
                "is_markup": None,
                "rendered_code": None,
            }
        else:
            readme_filename = force_unicode(readme_filename)
            readme_data = force_unicode(readme_data)
            return {
                "is_markup": markup.can_render(readme_filename),
                "rendered_code": highlight_or_render(readme_data, readme_filename),
            }


class HistoryView(TreeViewMixin, ReadmeMixin, BaseRepoView):
    """Show commits of a branch + path, just like `git log`. With pagination.

    Also, README, if available."""

    template_name = "history.html"

    def make_template_context(self, *args):
        super().make_template_context(*args)

        try:
            page = int(request.args.get("page"))
        except (TypeError, ValueError):
            page = 0

        self.context["page"] = page

        history_length = 30
        if page:
            skip = (self.context["page"] - 1) * history_length + 10
            if page > 7:
                self.context["previous_pages"] = [0, 1, 2, None] + list(range(page))[
                    -3:
                ]
            else:
                self.context["previous_pages"] = range(page)
        else:
            skip = 0

        history = self.context["repo"].history(
            self.context["commit"], self.context["path"], history_length + 1, skip
        )
        if len(history) == history_length + 1:
            # At least one more commit for next page left
            more_commits = True
            # We don't want show the additional commit on this page
            history.pop()
        else:
            more_commits = False

        self.context.update(
            {
                "history": history,
                "more_commits": more_commits,
            }
        )

        self.context.update(self.get_readme_context())


class IndexView(TreeViewMixin, ReadmeMixin, BaseRepoView):
    """Show commits of a branch, just like `git log`.

    Also, README, if available."""

    template_name = "index.html"

    def make_template_context(self, *args):
        super().make_template_context(*args)

        self.context["base_href"] = url_for(
            "blob",
            repo=self.context["repo"].namespaced_name,
            rev=self.context["rev"],
            path="",
        )

        self.context["page"] = 0
        history_length = 10
        history = self.context["repo"].history(
            self.context["commit"],
            self.context["path"],
            history_length + 1,
            skip=0,
        )
        if len(history) == history_length + 1:
            # At least one more commit for next page left
            more_commits = True
            # We don't want show the additional commit on this page
            history.pop()
        else:
            more_commits = False

        self.context.update(
            {
                "history": history,
                "more_commits": more_commits,
            }
        )

        self.context.update(self.get_readme_context())


class BaseBlobView(BaseRepoView):
    def make_template_context(self, *args):
        super().make_template_context(*args)
        if not isinstance(self.context["blob_or_tree"], dulwich.objects.Blob):
            raise NotFound("Not a blob")
        self.context["filename"] = os.path.basename(self.context["path"])


class SubmoduleView(BaseRepoView):
    """Show an information page about a submodule."""

    template_name = "submodule.html"

    def make_template_context(self, repo, namespace, rev, path):
        repo, rev, path, commit = _get_repo_and_rev(repo, namespace, rev, path)

        try:
            submodule_rev = tree_lookup_path(
                repo.__getitem__, commit.tree, encode_for_git(path)
            )[1]
        except KeyError:
            raise NotFound("Parent path for submodule missing")

        try:
            (submodule_url, submodule_path) = _get_submodule(
                repo, commit, encode_for_git(path)
            )
        except KeyError:
            submodule_url = None
            submodule_path = None

        # TODO(jelmer): Rather than printing an information page,
        # redirect to the page in klaus for the repository at
        # submodule_path, revision submodule_rev.

        self.context = {
            "view": self.view_name,
            "repo": repo,
            "rev": rev,
            "commit": commit,
            "branches": repo.get_branch_names(exclude=rev),
            "tags": repo.get_tag_names(),
            "path": path,
            "subpaths": list(subpaths(path)) if path else None,
            "submodule_url": force_unicode(submodule_url),
            "submodule_path": force_unicode(submodule_path),
            "submodule_rev": force_unicode(submodule_rev),
            "base_href": None,
        }


class BaseFileView(TreeViewMixin, BaseBlobView):
    """Base for FileView and BlameView."""

    def render_code(self, render_markup):
        should_use_ctags = current_app.should_use_ctags(
            self.context["repo"], self.context["commit"]
        )
        if should_use_ctags:
            if ctags is None:
                raise ImportError("Ctags enabled but python-ctags not installed")
            ctags_base_url = url_for(
                self.view_name,
                repo=self.context["repo"].namespaced_name,
                rev=self.context["rev"],
                path="",
            )
            ctags_tagsfile = CTAGS_CACHE.get_tagsfile(
                self.context["repo"].path, self.context["commit"].id
            )
            ctags_args = {
                "ctags": ctags.CTags(
                    ctags_tagsfile.encode(sys.getfilesystemencoding())
                ),
                "ctags_baseurl": ctags_base_url,
            }
        else:
            ctags_args = {}

        return highlight_or_render(
            force_unicode(self.context["blob_or_tree"].data),
            self.context["filename"],
            render_markup,
            **ctags_args,
        )

    def make_template_context(self, *args):
        super().make_template_context(*args)
        self.context.update(
            {
                "can_render": True,
                "is_binary": False,
                "too_large": False,
                "is_markup": False,
            }
        )

        binary = guess_is_binary(self.context["blob_or_tree"])
        too_large = sum(map(len, self.context["blob_or_tree"].chunked)) > 100 * 1024
        if binary:
            self.context.update(
                {
                    "can_render": False,
                    "is_binary": True,
                    "is_image": guess_is_image(self.context["filename"]),
                }
            )
        elif too_large:
            self.context.update(
                {
                    "can_render": False,
                    "too_large": True,
                }
            )


class FileView(BaseFileView):
    """Shows a file rendered using ``pygmentize``."""

    template_name = "view_blob.html"

    def make_template_context(self, *args):
        super().make_template_context(*args)
        if self.context["can_render"]:
            render_markup = "markup" not in request.args
            self.context.update(
                {
                    "is_markup": markup.can_render(self.context["filename"]),
                    "render_markup": render_markup,
                    "rendered_code": self.render_code(render_markup),
                }
            )


class BlameView(BaseFileView):
    template_name = "blame_blob.html"

    def make_template_context(self, *args):
        super().make_template_context(*args)
        if self.context["can_render"]:
            line_commits = self.context["repo"].blame(
                self.context["commit"], self.context["path"]
            )
            replace_dupes(line_commits, None)
            self.context.update(
                {
                    "rendered_code": self.render_code(render_markup=False),
                    "line_commits": line_commits,
                }
            )


class RawView(BaseBlobView):
    """Show a single file in raw for (as if it were a normal filesystem file
    served through a static file server).
    """

    def get_response(self):
        # Explicitly set an empty mimetype. This should work well for most
        # browsers as they do file type recognition anyway.
        # The correct way would be to implement proper file type recognition here.
        return Response(self.context["blob_or_tree"].chunked, mimetype="")


class DownloadView(BaseRepoView):
    """Download a repo as a tar.gz file."""

    def get_response(self):
        basename = "{}@{}".format(
            self.context["repo"].name,
            sanitize_branch_name(self.context["rev"]),
        )
        tarname = basename + ".tar.gz"
        headers = {
            "Content-Disposition": "attachment; filename=%s" % tarname,
            "Cache-Control": "no-store",  # Disables browser caching
        }

        tar_stream = dulwich.archive.tar_stream(
            self.context["repo"],
            self.context["blob_or_tree"],
            self.context["commit"].commit_time,
            format="gz",
            prefix=encode_for_git(basename),
        )
        return Response(tar_stream, mimetype="application/x-tgz", headers=headers)


history = HistoryView.as_view("history", "history")
index = IndexView.as_view("index", "index")
commit = CommitView.as_view("commit", "commit")
patch = PatchView.as_view("patch", "patch")
blame = BlameView.as_view("blame", "blame")
blob = FileView.as_view("blob", "blob")
raw = RawView.as_view("raw", "raw")
download = DownloadView.as_view("download", "download")
submodule = SubmoduleView.as_view("submodule", "submodule")



import os

try:
    from importlib import reload  # Python 3.4+
except ImportError:
    pass

import subprocess

import pytest
import requests

from klaus.contrib import app_args

from .utils import *


def check_env(env, expected_args, expected_kwargs):
    os.environ.update(env)
    args, kwargs = app_args.get_args_from_env()
    assert args == expected_args
    assert kwargs == expected_kwargs


def test_minimum_env(monkeypatch):
    """Test to provide only required env var"""
    monkeypatch.setattr(os, "environ", os.environ.copy())
    check_env(
        {"KLAUS_SITE_NAME": TEST_SITE_NAME},
        ([], TEST_SITE_NAME),
        dict(
            htdigest_file=None,
            use_smarthttp=False,
            require_browser_auth=False,
            disable_push=False,
            unauthenticated_push=False,
            ctags_policy="none",
        ),
    )


def test_complete_env(monkeypatch):
    """Test to provide all supported env var"""
    monkeypatch.setattr(os, "environ", os.environ.copy())
    check_env(
        {
            "KLAUS_REPOS": TEST_REPO_NO_NAMESPACE,
            "KLAUS_SITE_NAME": TEST_SITE_NAME,
            "KLAUS_HTDIGEST_FILE": HTDIGEST_FILE,
            "KLAUS_USE_SMARTHTTP": "yes",
            "KLAUS_REQUIRE_BROWSER_AUTH": "1",
            "KLAUS_DISABLE_PUSH": "false",
            "KLAUS_UNAUTHENTICATED_PUSH": "0",
            "KLAUS_CTAGS_POLICY": "ALL",
        },
        ([TEST_REPO_NO_NAMESPACE], TEST_SITE_NAME),
        dict(
            htdigest_file=HTDIGEST_FILE,
            use_smarthttp=True,
            require_browser_auth=True,
            disable_push=False,
            unauthenticated_push=False,
            ctags_policy="ALL",
        ),
    )


def test_unsupported_boolean_env(monkeypatch):
    """Test that unsupported boolean env var raises ValueError"""
    monkeypatch.setattr(os, "environ", os.environ.copy())
    with pytest.raises(ValueError):
        check_env(
            {
                "KLAUS_REPOS": TEST_REPO_NO_NAMESPACE,
                "KLAUS_SITE_NAME": TEST_SITE_NAME,
                "KLAUS_HTDIGEST_FILE": HTDIGEST_FILE,
                "KLAUS_USE_SMARTHTTP": "unsupported",
            },
            (),
            {},
        )


def test_wsgi(monkeypatch):
    """Test start of wsgi app"""
    monkeypatch.setattr(os, "environ", os.environ.copy())
    os.environ["KLAUS_REPOS"] = TEST_REPO_NO_NAMESPACE
    os.environ["KLAUS_SITE_NAME"] = TEST_SITE_NAME
    from klaus.contrib import wsgi

    with serve_app(wsgi.application):
        assert can_reach_unauth()
        assert not can_push_auth()

    os.environ["KLAUS_HTDIGEST_FILE"] = HTDIGEST_FILE
    os.environ["KLAUS_USE_SMARTHTTP"] = "yes"
    reload(wsgi)
    with serve_app(wsgi.application):
        assert can_reach_unauth()
        assert can_push_auth()


def test_wsgi_autoreload(monkeypatch):
    """Test start of wsgi autoreload app"""
    monkeypatch.setattr(os, "environ", os.environ.copy())
    os.environ["KLAUS_REPOS_ROOT"] = TEST_REPO_NO_NAMESPACE_ROOT
    os.environ["KLAUS_SITE_NAME"] = TEST_SITE_NAME
    from klaus.contrib import wsgi_autoreload, wsgi_autoreloading

    with serve_app(wsgi_autoreload.application):
        assert can_reach_unauth()
        assert not can_push_auth()

    os.environ["KLAUS_HTDIGEST_FILE"] = HTDIGEST_FILE
    os.environ["KLAUS_USE_SMARTHTTP"] = "yes"
    reload(wsgi_autoreload)
    reload(wsgi_autoreloading)
    with serve_app(wsgi_autoreload.application):
        assert can_push_auth()


def can_reach_unauth():
    return _check_http200(_GET_unauth, TEST_REPO_NO_NAMESPACE_BASE_URL)


def can_push_auth():
    return _can_push(_GET_auth, AUTH_TEST_REPO_NO_NAMESPACE_URL)


def _can_push(http_get, url):
    return any(
        [
            _check_http200(
                http_get,
                TEST_REPO_NO_NAMESPACE_BASE_URL + "info/refs?service=git-receive-pack",
            ),
            _check_http200(
                http_get, TEST_REPO_NO_NAMESPACE_BASE_URL + "git-receive-pack"
            ),
            subprocess.call(["git", "push", url, "master"], cwd=TEST_REPO_NO_NAMESPACE)
            == 0,
        ]
    )


def _GET_unauth(url=""):
    return requests.get(
        UNAUTH_TEST_SERVER + url,
        auth=requests.auth.HTTPDigestAuth("invalid", "password"),
    )


def _GET_auth(url=""):
    return requests.get(
        AUTH_TEST_SERVER + url,
        auth=requests.auth.HTTPDigestAuth("testuser", "testpassword"),
    )


def _check_http200(http_get, url):
    return http_get(url).status_code == 200



import os
import shutil
import subprocess
import tempfile


def check_have_compatible_ctags():
    """Check that the 'ctags' binary is a compatible ctags (Universal or Exuberant, not etags etc)"""
    try:
        out = subprocess.check_output(["ctags", "--version"], stderr=subprocess.PIPE)
        return b"Universal" in out or b"Exuberant" in out
    except subprocess.CalledProcessError:
        return False


def create_tagsfile(git_repo_path, git_rev):
    """Create a ctags tagsfile for the given Git repository and revision.

    This creates a temporary clone of the repository, checks out the revision,
    runs 'ctags -R' and deletes the temporary clone.

    :return: path to the generated tagsfile
    """
    assert (
        check_have_compatible_ctags()
    ), "'ctags' binary is missing or not *Universal* (or *Exuberant*) ctags"

    _, target_tagsfile = tempfile.mkstemp()
    checkout_tmpdir = tempfile.mkdtemp()
    try:
        subprocess.check_call(
            ["git", "clone", "-q", "--shared", git_repo_path, checkout_tmpdir]
        )
        subprocess.check_call(["git", "checkout", "-q", git_rev], cwd=checkout_tmpdir)
        subprocess.check_call(
            ["ctags", "--fields=+l", "-Rno", target_tagsfile], cwd=checkout_tmpdir
        )
    finally:
        shutil.rmtree(checkout_tmpdir)
    return target_tagsfile


def delete_tagsfile(tagsfile_path):
    """Delete a tagsfile."""
    os.remove(tagsfile_path)



from klaus import make_app

from .app_args import get_args_from_env

args, kwargs = get_args_from_env()

if kwargs["htdigest_file"]:
    with open(kwargs["htdigest_file"]) as file:
        kwargs["htdigest_file"] = file
        application = make_app(*args, **kwargs)
else:
    application = make_app(*args, **kwargs)



import argparse
import os
import sys
import webbrowser

from dulwich.errors import NotGitRepository
from dulwich.repo import Repo

from klaus import KLAUS_VERSION, make_app
from klaus.utils import force_unicode


def git_repository(path):
    path = os.path.abspath(path)
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError("%r: No such directory" % path)
    try:
        Repo(path)
    except NotGitRepository:
        raise argparse.ArgumentTypeError("%r: Not a Git repository" % path)
    return path


def make_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", help="default: 127.0.0.1", default="127.0.0.1")
    parser.add_argument("--port", help="default: 8080", default=8080, type=int)
    parser.add_argument(
        "--site-name", help="site name showed in header. default: your hostname"
    )
    parser.add_argument("--version", help="print version number", action="store_true")
    parser.add_argument(
        "-b",
        "--browser",
        help="open klaus in a browser on server start",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-B",
        "--with-browser",
        help="specify which browser to use with --browser",
        metavar="BROWSER",
        default=None,
    )
    parser.add_argument(
        "--ctags",
        help="enable ctags for which revisions? default: none. "
        "WARNING: Don't use 'ALL' for public servers!",
        choices=["none", "tags-and-branches", "ALL"],
        default="none",
    )

    parser.add_argument(
        "repos",
        help="repositories to serve",
        metavar="DIR",
        nargs="*",
        type=git_repository,
    )

    grp = parser.add_argument_group("Git Smart HTTP")
    grp.add_argument(
        "--smarthttp", help="enable Git Smart HTTP serving", action="store_true"
    )
    grp.add_argument(
        "--htdigest",
        help="use credentials from FILE",
        metavar="FILE",
        type=argparse.FileType("r"),
    )

    grp = parser.add_argument_group("Development flags", "DO NOT USE IN PRODUCTION!")
    grp.add_argument(
        "--debug", help="Enable Werkzeug debugger and reloader", action="store_true"
    )

    return parser


def main():
    args = make_parser().parse_args()

    if args.version:
        print(KLAUS_VERSION)
        return 0

    if args.htdigest and not args.smarthttp:
        print(
            "ERROR: --htdigest option has no effect without --smarthttp enabled",
            file=sys.stderr,
        )
        return 1

    if not args.repos:
        print(
            "WARNING: No repositories supplied -- syntax is 'klaus dir1 dir2...'.",
            file=sys.stderr,
        )

    if not args.site_name:
        args.site_name = "%s:%d" % (args.host, args.port)

    if args.ctags != "none":
        from klaus.ctagsutils import check_have_compatible_ctags

        if not check_have_compatible_ctags():
            print(
                "ERROR: Exuberant ctags not installed (or 'ctags' binary isn't *Exuberant* ctags)",
                file=sys.stderr,
            )
            return 1
        try:
            pass
        except ImportError:
            raise ImportError("Please install 'python-ctags3' to enable ctags support.")

    app = make_app(
        args.repos,
        force_unicode(args.site_name or args.host),
        args.smarthttp,
        args.htdigest,
        ctags_policy=args.ctags,
    )

    if args.browser:
        _open_browser(args)

    app.run(args.host, args.port, args.debug)


def _open_browser(args):
    # Open a web browser onto the server URL. Technically we're jumping the
    # gun a little here since the server is not yet running, but there's no
    # clean way to run a function after the server has started without
    # losing the simplicity of the code. In the Real World (TM) it'll take
    # longer for the browser to start than it will for us to start
    # serving, so we'll be OK.
    if args.with_browser is None:
        opener = webbrowser.open
    else:
        opener = webbrowser.get(args.with_browser).open
    opener(f"http://{args.host}:{args.port}")


if __name__ == "__main__":
    exit(main())



import os

from setuptools import setup

long_description = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()


def install_data_files_hack():
    # This is a clever hack to circumvent distutil's data_files
    # policy "install once, find never". Definitely a TODO!
    # -- https://groups.google.com/group/comp.lang.python/msg/2105ee4d9e8042cb
    from distutils.command.install import INSTALL_SCHEMES

    for scheme in INSTALL_SCHEMES.values():
        scheme["data"] = scheme["purelib"]


install_data_files_hack()

requires = [
    "flask",
    "Werkzeug>=0.15.0",
    "pygments",
    "httpauth>=0.4",
    "humanize",
    "dulwich>=0.19.3",
]

setup(
    name="klaus",
    version="3.0.1",
    author="Jonas Haag",
    author_email="jonas@lophus.org",
    packages=["klaus", "klaus.contrib"],
    include_package_data=True,
    zip_safe=False,
    url="https://github.com/jonashaag/klaus",
    description="The first Git web viewer that Just Works™.",
    long_description=long_description,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Version Control",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
    install_requires=requires,
    entry_points={
        "console_scripts": ["klaus=klaus.cli:main"],
    },
)



"""
    lodgeit.lib.diff
    ~~~~~~~~~~~~~~~~

    Render a nice diff between two things.

    :copyright: 2007 by Armin Ronacher.
    :license: BSD
"""

from difflib import SequenceMatcher

from klaus.utils import escape_html as e


def highlight_line(old_line, new_line):
    """Highlight inline changes in both lines."""
    start = 0
    limit = min(len(old_line), len(new_line))
    while start < limit and old_line[start] == new_line[start]:
        start += 1
    end = -1
    limit -= start
    while -end <= limit and old_line[end] == new_line[end]:
        end -= 1
    end += 1
    if start or end:

        def do(line, tag):
            last = end + len(line)
            return b"".join(
                [
                    line[:start],
                    b"<",
                    tag,
                    b">",
                    line[start:last],
                    b"</",
                    tag,
                    b">",
                    line[last:],
                ]
            )

        old_line = do(old_line, b"del")
        new_line = do(new_line, b"ins")
    return old_line, new_line


def render_diff(a, b, n=3):
    """Parse the diff an return data for the template."""
    actions = []
    chunks = []
    for group in SequenceMatcher(None, a, b).get_grouped_opcodes(n):
        lines = []

        def add_line(old_lineno, new_lineno, action, line):
            actions.append(action)
            lines.append(
                {
                    "old_lineno": old_lineno,
                    "new_lineno": new_lineno,
                    "action": action,
                    "line": line,
                    "no_newline": not line.endswith(b"\n"),
                }
            )

        chunks.append(lines)
        for tag, i1, i2, j1, j2 in group:
            if tag == "equal":
                for c, line in enumerate(a[i1:i2]):
                    add_line(i1 + c, j1 + c, "unmod", e(line))
            elif tag == "insert":
                for c, line in enumerate(b[j1:j2]):
                    add_line(None, j1 + c, "add", e(line))
            elif tag == "delete":
                for c, line in enumerate(a[i1:i2]):
                    add_line(i1 + c, None, "del", e(line))
            elif tag == "replace":
                for c, line in enumerate(a[i1:i2]):
                    add_line(i1 + c, None, "del", e(line))
                for c, line in enumerate(b[j1:j2]):
                    add_line(None, j1 + c, "add", e(line))
            else:
                raise AssertionError("unknown tag %s" % tag)

    return actions.count("add"), actions.count("del"), chunks
